/*
	Vector - the Vector class has been designed to complement
		 my other container classes, Stack, Queue and
		 DoubleList. It simply creates a Vector of a
		 requested size that can be used to store pointers
		 to object passed to it via put, you can look at
		 the pointer via peekAt or use get which NULLifies
		 the index within the Vector.

		 This class was initally designed for pointers to
		 any object type, but if you wish to use non-pointers
		 i.e. constants or built in types you will need to
		 ensure that the class does not attempt to delete these
		 types, the compiler should complain or unpredicatable
		 results will be forth coming. To stop the elements
		 of the Vector being deleted just insert the #define
		 VECTOR_NO_DELETE, obviously the vector/array itself
		 will be deleted at the end but no elements within it
		 will be. If you are using pointers and require the
		 delete feature turned off just use the run-time
		 function 'destroy'.
*/

#ifndef __VECTOR_H
#define __VECTOR_H

#ifdef DEBUG_ALL
#define DEBUG_VECTOR
#endif

#ifdef _Windows
 #undef DEBUG_ALL
 #undef DEBUG_VECTOR
#endif

#ifdef DEBUG_VECTOR
 #ifndef __ALLOC_H
  #include <alloc.h>
 #endif
 #ifndef __IOSTREAM_H
  #include <iostream.h>
 #endif
#endif

class	VectorError {
	public:
	enum errFlags {
		clear	  = 0x00,
		memory 	  = 0x01,
		range  	  = 0x02,
		overwrite = 0x04,
		error  	  = memory |
			    range |
			    overwrite
	};
};

template< class T > class Vector {
			 T*	aType;
		unsigned int	state;
		unsigned int	theSize;
		unsigned int	doDestroy;
			 int	fill;

		void	flushAll();
	public:
			Vector( unsigned int sz, int fill = 0 );
			Vector( Vector< T >& v );

	virtual		~Vector();

	virtual	const T	peekAt( unsigned int idx );
	virtual	const T	put( T type, unsigned int idx );
	virtual	const T	get( unsigned int idx );

	virtual void	flush();
		int	sizeOf() const
				{ return theSize; }
		int	isEmpty() const
				{ return theSize == 0; }
		int	destroy( unsigned int d )
				{ return doDestroy = d; }

		int	rdstate()
				{ return state; }
		int	setf( unsigned int s );
		void	clear( unsigned int s = VectorError::error );

	virtual	int	isA() const
				{ return 0; }
	virtual	char*	nameOf() const
				{ return "VectorClass"; }

	virtual	int	doEach( T type );
	virtual	int	forEach( int(*pfEach)( T ), int startAt = 0 );
	virtual	int	forEach( int startAt = 0 );

	virtual	void	outOfRange();
	virtual	void	outOfMemory();
	virtual	void	overWrite( unsigned int idx );

		T&	operator[]( unsigned int idx );
	   Vector< T >& operator=( Vector< T >& v );
		int	operator!();
			operator Vector< T >*();
};

/*
	purpose: constructor, accepts a size and inits, the array
		 to size sz and clears each element to zero. If
		 memory is unable to be allocated then calls virtual
		 of outOfMemory.
*/

template< class T > Vector< T >::Vector( unsigned int sz, int f )
{
	#ifdef DEBUG_VECTOR
		cout << "Memory at "
		     << nameOf()
		     << " constructor: "
		     << farcoreleft() << endl;
	#endif

	theSize = 0;
	state   = 0;
	doDestroy = 1;
	aType   = new T[sz];

	if( !aType ) {
		outOfMemory();
		return;
	}

	theSize = sz;
	fill    = f;

	for( int i = 0; i < theSize; i++ )
		aType[i] = (T)fill;
}

/*
	purpose: copy constructor, see above.
*/

template< class T > Vector< T >::Vector( Vector< T >& v )
{
	#ifdef DEBUG_VECTOR
		cout << "Memory at "
		     << nameOf()
		     << " constructor: "
		     << farcoreleft() << endl;
	#endif
	state = 0;
	aType = 0;
	*this = v;
}

/*
	purpose: destructor, flushes all the elements.
*/

template< class T > Vector< T >::~Vector()
{
	flushAll();

	#ifdef DEBUG_VECTOR
		cout << "Memory at "
		     << nameOf()
		     << " destructor: "
		     << farcoreleft() << endl;
	#endif

}

/*
	purpose: used to clear specific flags, i.e. clear error->
		 clear( VectorError::error ).

	returns: void
*/

template< class T > void Vector< T >::clear( unsigned int s )
{
	state ^= (state & s);
}

/*
	purpose: to set the state flag to the users request,
		 returns the state prior to the user changing it.
*/

template< class T > int Vector< T >::setf( unsigned int s )
{
unsigned int copy = state;

	state = s;
	return copy;
}

/*
	purpose: to use the flush function to clear all elements of
		 the Vector and to also clear down the vector itself.
*/

template< class T > void Vector< T >::flushAll()
{
	flush();
	delete[] aType;
	aType = 0;
}

/*
	purpose: flush simply deletes all the elements within
		 the Vector, whilst leaving the Vector itself
		 int tact.
*/

template< class T > void Vector< T >::flush()
{
      if( theSize && doDestroy ) {
	for( int i = 0; i < theSize; i++ ) {
	  #ifndef VECTOR_NO_DELETE
	     if( aType[i] )
		delete aType[i];
	  #endif
	     aType[i] = (T)fill;
	}
       }
}

/*
	purpose: to be derived from and overriden. If an index greater
		 than the current size if attempted outOfRange is called.
*/

template< class T > void Vector< T >::outOfRange()
{
	state |= VectorError::range;
}

template< class T > void Vector< T >::outOfMemory()
{
	state |= VectorError::memory;
}

/*
	purpose: if you try to put an item into an index that
		 already contains data, it is classed as an
		 overwrite error. If doDestroy is set the item
		 is deleted.
*/

#ifdef VECTOR_NO_DELETE
#pragma warn -par
#endif
template< class T > void Vector< T >::overWrite( unsigned int idx )
{
	state |= VectorError::overwrite;
	#ifndef VECTOR_NO_DELETE
		if( doDestroy ) delete aType[idx];
	#endif
}

/*
	purpose: to allow a simple way to index through the Vector
		 carrying out operations on each type.
		 NOTE: it is the users responsibility to ensure
		 an object exists, i.e this may pass a NULL pointer
		 to your 'theFunction' so it is your responsibility
		 to test for this.
		 To stop forEach running prior to reaching the end of
		 the Vector just return 0 from 'theFunction'.
*/

template< class T > int Vector< T >::forEach( int (*theFunction)( T ), int startAt )
{
	for( int i = startAt; i < theSize; i++ )
		if( !theFunction( (T)peekAt( i ) ) )
			break;

	return i + 1;
}

template< class T > int Vector< T >::forEach( int startAt )
{
	for( int i = startAt; startAt < theSize; i++ )
		if( !doEach( peekAt( i ) ) )
			break;

	return i + 1;
}

template< class T > int Vector< T >::doEach( T )
{
	return 1;
}

template< class T > const T Vector< T >::peekAt( unsigned int idx )
{
	if( idx >= theSize ) {
		outOfRange();
		return 0;
	}

	return aType[idx];
}

template< class T > const T Vector< T >::put( T type, unsigned int idx )
{
	if( idx >= theSize ) {
		outOfRange();
		#ifndef VECTOR_NO_DELETE
		    if( doDestroy )
			delete type;
		#endif
		return 0;
	}

	// if something exists at the index, defined by the fill
	// type.
	if( aType[idx] != (T)fill )
		overWrite( idx );

	return aType[idx] = type;
}

template< class T > const T Vector< T >::get( unsigned int idx )
{
T	type;

	if( idx >= theSize ) {
		outOfRange();
		return 0;
	}

	type = aType[idx];
	aType[idx] = (T)fill;
	return type;
}

/*
	purpose: to emulate built in type arrays, note: the error
		 handling is non existant, i.e if the index is greater
		 than the available size the last element is returned.
*/

template< class T > T& Vector< T >::operator[]( unsigned int idx )
{
	// note: this is an awful way of returning a reference error
	// let's hope for exception handling...
	if( idx >= theSize ) {
		outOfRange();
		return aType[theSize - 1];
	}

	return aType[idx];
}

template< class T > Vector< T >& Vector< T >::operator=( Vector< T >& v )
{
	if( aType ) flushAll();

	state = 0;
	theSize = 0;
	doDestroy = 1;

	aType = new T[v.sizeOf()];
	if( !aType ) {
		outOfMemory();
		return *this;
	}

	theSize = v.sizeOf();

	for( int i = 0; i < theSize; i++ )
		if( !put( v.peekAt( i ), i ) )
			break;

	return *this;
}

template< class T > int Vector< T >::operator!()
{
	return state & VectorError::error;
}

template< class T > Vector< T >::operator Vector< T >*()
{
	return (state & VectorError::error) ? 0 : this;
}

#endif
